import numpy as np
import pandas as pd
from .Concept import Concept


class Dataset():
    """Wrapper for input datasets, resulting in one or more DDF tables.

    Parameters
    ----------
    data : pandas.DataFrame
        The data to packaged.
    concepts : list
        List of dictionaries each with keys concept and
        concept_type for every concept in the dataset.
    """
    entities_metadata = {}

    def __init__(self, data, concepts):
        if type(data) is not pd.DataFrame:
            msg = (f'Expected data to be of type pandas.DataFrame'
                   f', not {type(data)}')
            raise TypeError(msg)

        if type(concepts) is not list:
            msg = (f'Expected concepts to be of type list'
                   f', not {type(concepts)}')
            raise TypeError(msg)

        if not all([x in [y['concept'] for y in concepts]
                    for x in data.columns]):
            raise ValueError('Missing concepts.')
        self.data = data
        self.concepts = [Concept(x) for x in concepts]

    @property
    def entities(self):
        ent_map = {}
        entities = filter(lambda x: x.is_entity, self.concepts)

        for ent in entities:
            name = ent.name
            values = self.data[[name]]
            values = values.drop_duplicates()
            values = values.dropna(subset=[name])
            values = values.applymap(str.lower)

            if name in self.entities_metadata:
                metadata = self.entities_metadata[name]
                values = values.merge(metadata, on=name)

            ent_map[name] = values

        return ent_map

    def update_entity(self, name, data):
        if name not in self.entities:
            raise ValueError(f'Entity {name} does not exist')

        self.entities_metadata[name] = data
        for col in data.columns:
            if col in self.concepts:
                continue
            concept = {'concept': col, 'concept_type': 'string'}
            self.concepts.append(Concept(concept))

    def _filter_concepts(self, types):
        filt_func = lambda x: x.type in types
        return filter(filt_func, self.concepts)

    @property
    def tables(self):
        tables = []
        measures = self._filter_concepts(['measure'])
        pks = self._filter_concepts(['time', 'entity_domain'])

        for m in measures:
            fname = f'ddf--datapoints--{m.name}'
            cols = [x.name for x in pks] + [m.name]
            table = self.data[cols]
            table = table.replace('', np.nan)
            table = table.dropna(subset=[m.name])
            table = table.dropna(how='all', axis=1)
            table = table.applymap(lambda x: x.lower() if type(x) is str else x)

            # Missing primary keys interpreted as summaries
            # so we add extra tables
            if table.isnull().any().any():
                null_cols = list(table.columns[table.isnull().any()])
                for nc in null_cols:
                    nct = table[table[nc].isnull()]
                    nct = nct.drop(nc, axis=1)
                    pk = "--".join(sorted(nct.columns.drop(m.name)))
                    fnamenc = f'{fname}--by--{pk}.csv'
                    tables.append((fnamenc, nct))

                table = table.dropna(subset=null_cols)

            pk = "--".join(sorted(table.columns.drop(m.name)))
            fname = f'{fname}--by--{pk}.csv'

            if len(table) > 0:
                tables.append((fname, table))

        return tables

    def __repr__(self):
        return f'<Dataset: {list(self.data.columns)}>'



import os
import shutil
import pandas as pd
import numpy as np
from ddf_utils import package
from ddf_utils.io import dump_json
from .Dataset import Dataset
from .Concept import Concept


class Frame2Package():
    """ Main class of frame2package."""

    def __init__(self):
        self.data = []

    def add_data(self, data, concepts):
        dataset = Dataset(data, concepts)
        self.data.append(dataset)

    def update_entity(self, name, data):
        for dataset in self.data:
            if name in dataset.entities:
                dataset.update_entity(name, data)

    def from_package(self, path):
        raise NotImplementedError

    @property
    def concepts(self):
        return [y for x in self.data for y in x.concepts]

    @property
    def entities(self):
        entities = {}
        for dataset in self.data:
            entities.update(dataset.entities)
        return entities

    def to_package(self, dirname, attrs={}):
        """ Save data to a DDF package.

        Parameters
        ----------
        dirname : str
            Name of the DDF directory to be created.
        attrs : dict
            Attributes to add/update in datapackage.json.
        """

        cwd = os.getcwd()
        dirpath = os.path.join(cwd, dirname)

        if os.path.exists(dirpath):
            shutil.rmtree(dirpath)

        os.mkdir(dirpath)

        # Create data files
        for dataset in self.data:
            for table in dataset.tables:
                path = os.path.join(dirpath, table[0])
                table[1].to_csv(path, index=False)

        # Create entity files
        for entity_name, entity_items in self.entities.items():
            path = os.path.join(dirpath, f'ddf--entities--{entity_name}.csv')
            entity_items.to_csv(path, index=False)

        # Create concepts file
        path = os.path.join(dirpath, 'ddf--concepts.csv')
        concepts = pd.DataFrame([x.data for x in self.concepts])
        concepts = concepts.drop_duplicates(subset=['concept'])
        concepts.to_csv(path, index=False)

        # Create datapackage.json
        meta = package.create_datapackage(dirpath, **attrs)
        dump_json(os.path.join(dirpath, 'datapackage.json'), meta)

    def __repr__(self):
        return (f'<Frame2Package: {len(self.data)} datasets, '
                f'{len(self.concepts)} concepts, '
                f'{len(self.entities)} entities>')



import unittest
from frame2package import Concept


class ConceptTestCase(unittest.TestCase):
    def setUp(self):
        concept = {
            'concept': 'country',
            'concept_type': 'entity_domain',
            'description': 'Name of country'
        }
        self.concept = Concept(concept)

    def test_has_name(self):
        self.assertTrue(hasattr(self.concept, 'name'))

    def test_has_correct_name(self):
        self.assertEqual(self.concept.name, 'country')

    def test_has_type(self):
        self.assertTrue(hasattr(self.concept, 'type'))

    def test_has_correct_type(self):
        self.assertEqual(self.concept.type, 'entity_domain')

    def test_if_entity(self):
        self.assertTrue(self.concept.is_entity)

    def test_equality(self):
        self.assertEqual(self.concept, 'country')

    def test_membership(self):
        self.assertIn('country', [self.concept])



class Concept():
    """Simple wrapper around concept dicts.

    Parameters
    ----------
    data : dict
        Dictionary with mandatory keys `concept` and `concept_type`.
    """

    def __init__(self, data):
        self.data = data

    @property
    def name(self):
        return self.data['concept']

    @property
    def type(self):
        return self.data['concept_type']

    @property
    def is_entity(self):
        return self.data['concept_type'] == 'entity_domain'

    def __eq__(self, value):
        return self.name == value

    def __repr__(self):
        name = self.data['concept']
        kind = self.data['concept_type']
        return f'<Concept: {name} ({kind})>'



import unittest
import time
import os
import shutil
import pandas as pd
from frame2package import Frame2Package


class Frame2PackageIoTestCase(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        data = [
            {
                'country': 'Sweden',
                'capital': 'Stockholm',
                'year': 2000,
                'population': 9_000_000
            },
            {
                'country': 'Sweden',
                'capital': 'Stockholm',
                'year': 2019,
                'population': 10_000_000
            },
            {
                'country': 'Norway',
                'capital': 'Oslo',
                'year': 2000,
                'population': 5_000_000
            },
            {
                'country': 'Norway',
                'capital': 'Oslo',
                'year': 2019,
                'population': 6_000_000
            },
        ]

        concepts = [
            {
                'concept': 'country',
                'concept_type': 'entity_domain'
            },
            {
                'concept': 'capital',
                'concept_type': 'string'
            },
            {
                'concept': 'population',
                'concept_type': 'measure'
            },
            {
                'concept': 'year',
                'concept_type': 'time'
            }
        ]

        f2p = Frame2Package()
        f2p.add_data(pd.DataFrame(data), concepts)
        path = f'ddf-package-test-{str(int(time.time()))}'
        f2p.to_package(path)

        self.f2p = f2p
        self.path = path

    def test_package_was_created(self):
        self.assertTrue(os.path.exists(self.path))

    def test_concepts_csv_was_created(self):
        concepts_path = os.path.join(self.path, 'ddf--concepts.csv')
        self.assertTrue(os.path.exists(concepts_path))

    def test_datapackage_json_was_created(self):
        datapackage_path = os.path.join(self.path, 'datapackage.json')
        self.assertTrue(os.path.exists(datapackage_path))

    def test_entities_csv_was_created(self):
        country_path = os.path.join(self.path, 'ddf--entities--country.csv')
        self.assertTrue(os.path.exists(country_path))

    @classmethod
    def tearDownClass(self):
        shutil.rmtree(self.path)


class Frame2PackageDataTestCase(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        data = [
            {
                'country': 'Sweden',
                'region': 'Nordic',
                'population': 9_000_000
            },
            {
                'country': 'Sweden',
                'region': 'Nordic',
                'population': 10_000_000
            },
            {
                'country': 'Denmark',
                'region': 'Nordic',
                'population': 5_000_000
            },
            {
                'country': None,
                'region': 'Nordic',
                'population': 50_000_000
            },
        ]

        concepts = [
            {
                'concept': 'country',
                'concept_type': 'entity_domain'
            },
            {
                'concept': 'region',
                'concept_type': 'entity_domain'
            },
            {
                'concept': 'population',
                'concept_type': 'measure'
            }
        ]

        f2p = Frame2Package()
        f2p.add_data(pd.DataFrame(data), concepts)
        self.f2p = f2p

    def test_disaggregation_levels(self):
        fnames = [x[0] for x in self.f2p.data[0].tables]

        country_region = 'ddf--datapoints--population--by--country--region.csv'
        self.assertIn(country_region, fnames)

        region = 'ddf--datapoints--population--by--region.csv'
        self.assertIn(region, fnames)



import unittest
import pandas as pd
from frame2package import Dataset, Concept


class DatasetTestCase(unittest.TestCase):
    def setUp(self):
        data = [
            {
                'country': 'Sweden',
                'capital': 'Stockholm',
                'year': 2000,
                'population': 9_000_000
            },
            {
                'country': 'Sweden',
                'capital': 'Stockholm',
                'year': 2019,
                'population': 10_000_000
            },
            {
                'country': 'Norway',
                'capital': 'Oslo',
                'year': 2000,
                'population': 5_000_000
            },
            {
                'country': 'Norway',
                'capital': 'Oslo',
                'year': 2019,
                'population': 6_000_000
            },
        ]

        concepts = [
            {
                'concept': 'country',
                'concept_type': 'entity_domain'
            },
            {
                'concept': 'capital',
                'concept_type': 'string'
            },
            {
                'concept': 'population',
                'concept_type': 'measure'
            },
            {
                'concept': 'year',
                'concept_type': 'time'
            }
        ]

        self.data = data
        self.concepts = concepts
        self.dataset = Dataset(pd.DataFrame(data), concepts)

    def test_has_concepts(self):
        self.assertTrue(hasattr(self.dataset, 'concepts'))

    def test_has_entities(self):
        self.assertTrue(hasattr(self.dataset, 'entities'))

    def test_has_tables(self):
        self.assertTrue(hasattr(self.dataset, 'tables'))

    def test_has_data(self):
        self.assertTrue(hasattr(self.dataset, 'data'))

    def test_data_is_frame(self):
        self.assertTrue(type(self.dataset.data) is pd.DataFrame)

    def test_concept_type(self):
        self.assertTrue(all([type(x) is Concept
                        for x in self.dataset.concepts]))

    def test_has_correct_number_of_entities(self):
        self.assertEqual(len(self.dataset.entities), 1)

    def test_fails_if_missing_concepts(self):
        data = pd.DataFrame(self.data)

        def create_dataset_with_missing_concepts():
            return Dataset(data, self.concepts[:-1])

        self.assertRaises(ValueError, create_dataset_with_missing_concepts)

    def test_creates_correct_table_name(self):
        table_name = self.dataset.tables[0][0]
        expected = 'ddf--datapoints--population--by--country--year.csv'
        self.assertEqual(table_name, expected)

    def test_creates_correct_table_size(self):
        self.assertEqual(self.dataset.tables[0][1].shape, (4, 3))

    def test_records_extra_string_concepts(self):
        self.assertIn('capital', self.dataset.concepts)



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="frame2package",
    version="0.0.4",
    author="Robin Linderborg",
    author_email="robin@datastory.org",
    description="Pandas dataframes to DDF packages.",
    download_url="https://github.com/datastory-org/frame2package/archive/0.0.4.tar.gz",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=setuptools.find_packages(),
    install_requires=[
        'pandas>=0.23',
        'ddf_utils==0.6.4'
     ],
)